# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/stable/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import re
import sys

package_path = os.path.abspath('../')
dhparser_path = os.path.abspath('../DHParser')
if package_path not in sys.path:
    sys.path.insert(0, package_path)
if dhparser_path not in sys.path:
    sys.path.insert(0, dhparser_path)
# sys.path.insert(0, os.path.abspath('../DHParser/scripts'))


# -- Project information -----------------------------------------------------

project = 'DHParser'
copyright = '2018-2021, Eckhart Arnold'
author = 'Eckhart Arnold'

# The short X.Y version
from DHParser import __version__
version = re.match('\d+\.\d+(?:\.\d+)?', __version__).group(0)
# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx.ext.napoleon'
    # 'sphinx_autodoc_typehints'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# A boolean that decides whether module names are prepended to all object names
add_module_names = False

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'  # 'alabaster'  # 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {
#     'body_max_width': '880px',
#     # 'documentwidth' : '1024px',
#     'page_width': '1024px',
#     'code_font_size': '0.70em'
# }

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['html_assets']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Patch to suppress type annotations --------------------------------------


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'DHParserdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'DHParser.tex', 'DHParser Documentation',
     'Eckhart Arnold', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'dhparser', 'DHParser Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'DHParser', 'DHParser Documentation',
     author, 'DHParser', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/3': None}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# Do not substitute type aliases by full type expression!
autodoc_type_aliases = {
  "ApplyFunc": "ApplyFunc",
  "BytesType": "BytesType",
  "CallItem": "CallItem",
  "ChildrenType": "ChildrenType",
  "CompilerCallable": "CompilerCallable",
  "CompilerFactory": "CompilerFactory",
  "CompilerFactoryFunc": "CompilerFactoryFunc",
  "CompilerFunc": "CompilerFunc",
  "ConditionFunc": "ConditionFunc",
  "ConnectionCallback": "ConnectionCallback",
  "CriteriaType": "CriteriaType",
  "CustomParseFunc": "CustomParseFunc",
  "DeriveFileNameFunc": "DeriveFileNameFunc",
  "ErrorMessagesType": "ErrorMessagesType",
  "FindIncludeFunc": "FindIncludeFunc",
  "FlagFunc": "FlagFunc",
  "JSON_Dict": "JSON_Dict",
  "JSON_Type": "JSON_Type",
  "Junction": "Junction",
  "KeyFunc": "KeyFunc",
  "MatchVariableFunc": "MatchVariableFunc",
  "MemoizationDict": "MemoizationDict",
  "NodeMatchFunction": "NodeMatchFunction",
  "NodeSelector": "NodeSelector",
  "ParseFunc": "ParseFunc",
  "ParserCallable": "ParserCallable",
  "ParserFactoryFunc": "ParserFactoryFunc",
  "ParserTrail": "ParserTrail",
  "ParsingResult": "ParsingResult",
  "Path": "Path",
  "PathMatchFunction": "PathMatchFunction",
  "PathSelector": "PathSelector",
  "PatternMatchType": "PatternMatchType",
  "PickChildFunction": "PickChildFunction",
  "PreprocessorFactoryFunc": "PreprocessorFactoryFunc",
  "PreprocessorFunc": "PreprocessorFunc",
  "RPC_Error_Type": "RPC_Error_Type",
  "RPC_Table": "RPC_Table",
  "RPC_Type": "RPC_Type",
  "ReentryPointAlgorithm": "ReentryPointAlgorithm",
  "ReprType": "ReprType",
  "ResultType": "ResultType",
  "ResumeList": "ResumeList",
  "Schema": "Schema",
  "SchemaLanguage": "SchemaLanguage",
  "SourceMapFunc": "SourceMapFunc",
  "StrictResultType": "StrictResultType",
  "SymbolsDictType": "SymbolsDictType",
  "Tokenizer": "Tokenizer",
  "TransformationCache": "TransformationCache",
  "TransformationDict": "TransformationDict",
  "TransformationProc": "TransformationProc",
  "TransformationTableType": "TransformationTableType",
  "TransformerCallable": "TransformerCallable",
  "TransformerFactoryFunc": "TransformerFactoryFunc",
  "ValidationFunction": "ValidationFunction",
  "re_pattern": "re_pattern",
}
